/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static java.lang.Math.min;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH;
import static org.fortiss.tooling.base.model.layout.EOrientation.EAST;
import static org.fortiss.tooling.base.model.layout.EOrientation.NORTH;
import static org.fortiss.tooling.base.model.layout.EOrientation.SOUTH;
import static org.fortiss.tooling.base.model.layout.EOrientation.WEST;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.getAbsolute;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.getConnectorOffsetOrientation;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.getNodePosition;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.setConnectorOffsetOrientation;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.setNodePosition;
import static org.fortiss.tooling.base.ui.utils.SnapToGridUtils.snapToGrid;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createOffsetOrientation;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.ui.dnd.gef.ElementDropContext;

/**
 * Utility class for rectangle-shaped edit part layout.
 * 
 * @author wandinger
 * @author hummel
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 5B8E9A8C0A5A0B34D1C7DD2A56368722
 */
public class RectangleLayoutUtils {

	/**
	 * Returns the {@link OffsetOrientation} of the point on the rectangle
	 * border having a given Point inside the rectangle.
	 * 
	 * @param source
	 *            the relative coordinates in the rectangle
	 * @param rectangleSize
	 *            the size of the rectangle
	 * @param inset
	 *            the inset from each side
	 * @return the offset and orientation
	 */
	public static OffsetOrientation calculateRectangleBorderOrientation(Point source,
			Dimension rectangleSize, int inset) {
		int distToNorth = source.y;
		int distToSouth = rectangleSize.height - source.y - 2 * inset;
		int distToWest = source.x;
		int distToEast = rectangleSize.width - source.x - 2 * inset;
		int minDist = min(min(min(distToEast, distToNorth), distToWest), distToSouth);

		int offset;
		EOrientation orientation;
		if(minDist == distToNorth) {
			orientation = NORTH;
			offset = source.x;
		} else if(minDist == distToWest) {
			orientation = WEST;
			offset = source.y;
		} else if(minDist == distToSouth) {
			orientation = SOUTH;
			offset = source.x;
		} else {
			orientation = EAST;
			offset = source.y;
		}
		return createOffsetOrientation(offset, orientation, null);
	}

	/**
	 * Ensure that the minimum size of a rectangular shape is redeemed.
	 * 
	 * @param oldSize
	 *            the old size to be adjusted
	 * @param minimumWidth
	 *            the minimum to be redeemed
	 * @param minimumHeight
	 *            the minimum height to be redeemed
	 * @return the new dimensions
	 */
	public static Dimension clampRectangularShapeSize(Dimension oldSize, int minimumWidth,
			int minimumHeight) {
		// check for minimum size
		Dimension newSize =
				new Dimension(oldSize.width < minimumWidth ? minimumWidth : oldSize.width,
						oldSize.height < minimumHeight ? minimumHeight : oldSize.height);
		return newSize;
	}

	/**
	 * Ensure that the minimum size of a rectangular shape is redeemed.
	 * 
	 * @param oldSize
	 *            the old size to be adjusted
	 * @return the new dimensions
	 */
	public static Dimension retainMinimumRectangularShapeSize(Dimension oldSize) {
		return clampRectangularShapeSize(oldSize, DEFAULT_SHAPE_MINIMUM_WIDTH,
				DEFAULT_SHAPE_MINIMUM_HEIGHT);
	}

	/**
	 * Layouts the given connector within the surrounding rectangle.
	 * 
	 * @param layouted
	 *            the connector element
	 * @param container
	 *            the container of the connector
	 * @param location
	 *            the location in the rectangle to be used as hint
	 */
	public static void layoutConnectorInRectangle(ILayoutedModelElement layouted,
			ILayoutedModelElement container, Point location) {
		Point position = location.getCopy();
		org.fortiss.tooling.base.model.layout.Point np = getNodePosition(container);
		position.translate(-np.getX() - DEFAULT_SHAPE_INSETS, -np.getY() - DEFAULT_SHAPE_INSETS);
		// do border clamping
		org.fortiss.tooling.base.model.layout.Dimension nd = getNodeSize(container);
		Dimension d = new Dimension(nd.getWidth(), nd.getHeight());
		OffsetOrientation orient =
				calculateRectangleBorderOrientation(position, d, DEFAULT_SHAPE_INSETS);
		// snap to grid and store the value in the model
		orient.setOffset(snapToGrid(orient.getOffset()));
		setConnectorOffsetOrientation(layouted, orient);
	}

	/**
	 * Layouts the given layouted object using the position in the context
	 * 
	 * @param layouted
	 *            the node to be layouted
	 * @param context
	 *            the location context information to be used
	 * @param positionOffset
	 *            the offset to adj ust the context location, e.g. grid size
	 * @param snapToGrid
	 *            snap to grid should be performed or not
	 */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted, ElementDropContext context,
			int positionOffset, boolean snapToGrid) {
		layoutNodeInRoot(layouted, context.getPosition(), positionOffset, snapToGrid);
	}

	/**
	 * Layouts the given layouted object using the position in the context.
	 * 
	 * @param layouted
	 *            the node to be layouted
	 * @param location
	 *            the location information to be used
	 * @param positionOffset
	 *            the offset to adjust the location, e.g. grid size
	 * @param snapToGrid
	 *            snap to grid should be performed or not
	 */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted, Point location,
			int positionOffset, boolean snapToGrid) {
		Point position = location.getCopy();
		position.translate(-positionOffset, -positionOffset);
		if(snapToGrid) {
			snapToGrid(position);
		}
		setNodePosition(layouted, position.x, position.y);
	}

	/**
	 * Returns the bounds rectangle of the connector given its parent bounds.
	 * 
	 * @param connector
	 *            the connector with relative positions
	 * @param parent
	 *            the parent container
	 * @returns the absolute rectangle
	 */
	public static Rectangle getAbsoluteConnectorBounds(ILayoutedModelElement connector,
			ILayoutedModelElement parent) {
		Rectangle result = new Rectangle(0, 0, DEFAULT_CONNECTOR_SIZE, DEFAULT_CONNECTOR_SIZE);
		// get orientation
		OffsetOrientation orient = getConnectorOffsetOrientation(connector);
		// get parent size
		org.fortiss.tooling.base.model.layout.Dimension parentSize = getNodeSize(parent);
		org.fortiss.tooling.base.model.layout.Point pos =
				getAbsolute(orient, parentSize, DEFAULT_SHAPE_INSETS);
		org.fortiss.tooling.base.model.layout.Point parentPos = getNodePosition(parent);
		int maxX = parentSize.getWidth() - DEFAULT_CONNECTOR_SIZE;
		int maxY = parentSize.getHeight() - DEFAULT_CONNECTOR_SIZE;
		result.x = parentPos.getX() + ((pos.getX() > maxX) ? maxX : pos.getX());
		result.y = parentPos.getY() + ((pos.getY() > maxY) ? maxY : pos.getY());
		return result;
	}
}
