/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2012 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.ui.picker;

import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.startPickingModeAndCountHits;
import static org.lwjgl.BufferUtils.createIntBuffer;

import java.nio.IntBuffer;
import java.util.HashMap;
import java.util.Map;

import org.fortiss.tooling.graphicsGL.ui.camera.Camera;
import org.fortiss.tooling.graphicsGL.ui.objects.ViewObjectBase;
import org.lwjgl.util.vector.Vector3f;

/**
 * Picker implementation for model objects using GL select mode rendering.
 * {@link #pickModelObject(int, int, Vector3f, ViewObjectBase, Camera)} returns the nearest object.
 * All others are discarded.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 4ECEE349CAECFD1FC8A9F4ECDB98E145
 */
public final class ModelObjectPicker {
	/** The buffer size for pickable objects. */
	private static final int BUFFER_SIZE = 1024;

	/** The selection buffer. */
	private final IntBuffer selectBuffer = createIntBuffer(BUFFER_SIZE);
	/** The viewport buffer. */
	private final IntBuffer viewport = createIntBuffer(16);

	/** The pickable object registry. */
	private final Map<Integer, Object> objectRegistry = new HashMap<Integer, Object>();

	/** The registry counter. */
	private int registryCounter = 1;

	/** Registers the given object for the current picker run. */
	public int registerPickableObject(Object object) {
		int name = registryCounter;
		objectRegistry.put(name, object);
		registryCounter++;
		return name;
	}

	/**
	 * Performs the picking process for the given {@link ViewObjectBase} from the given
	 * {@link Camera}s view at the given location and returns the picking result or {@code null} if
	 * nothing was hit.
	 */
	public Object
			pickModelObject(int x, int y, Vector3f perspective, ViewObjectBase scene, Camera c) {
		// reset picker
		objectRegistry.clear();
		registryCounter = 1;
		viewport.clear();
		selectBuffer.clear();

		int hits =
				startPickingModeAndCountHits(x, y, perspective, viewport, scene, this,
						selectBuffer, c);
		if(hits > 0) {
			int name = selectBuffer.get(3);
			int depth = selectBuffer.get(1);

			for(int i = 1; i < hits; i++) {
				int idx = i * 4;
				int nDepth = selectBuffer.get(idx + 1);
				if(nDepth < depth) {
					name = selectBuffer.get(idx + 3);
					depth = nDepth;
				}
			}
			if(name > 0) {
				return objectRegistry.get(name);
			}
		}
		return null;
	}
}
