/*--------------------------------------------------------------------------+
$Id: PersistencyService.java 5274 2012-08-02 07:54:11Z mou $
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.fortiss.tooling.kernel.extension.ILibraryElementHandler;
import org.fortiss.tooling.kernel.extension.IPrototypeProvider;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.ILibrary;
import org.fortiss.tooling.kernel.model.ILibraryElement;
import org.fortiss.tooling.kernel.model.ILibraryElementReference;
import org.fortiss.tooling.kernel.model.ILibraryPackage;
import org.fortiss.tooling.kernel.service.ILibraryService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.IPrototypeService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.utils.UniqueIDUtils;

/**
 * This class implements the {@link ILibraryService} interface.
 * 
 * @author ratiu
 * @author $Author: ratiu $
 * @version $Rev: 5274 $
 * @ConQAT.Rating GREEN Hash: 5EA401E2DCE151387F27B2FA71772A25
 */
public class LibraryService extends EObjectAwareServiceBase<ILibraryElementHandler<EObject>>
		implements ILibraryService {

	/** The library service extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.modelElementLibraryHandler";

	/** The library service configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelElementLibraryHandler";

	/** The handler name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "handler";

	/** {@inheritDoc} */
	@Override
	public boolean isLibraryElementShadow(EObject eobj) {
		ILibraryElementReference parent = getParentElement(eobj, ILibraryElementReference.class);

		// with "eobj.eContainer() != parent" we assume that the model elements are not directly
		// contained in the reference
		if(parent != null && parent != eobj) { // && eobj.eContainer() != parent) {
			return true;
		}

		return false;
	}

	/** Map holds for each {@link ILibraryElementReference} whether it is dirty or not. */
	private HashMap<ILibraryElementReference, Boolean> libRef2Dirty =
			new HashMap<ILibraryElementReference, Boolean>();

	/** {@inheritDoc} */
	@Override
	public boolean isDirty(ILibraryElementReference libRef) {
		if(libRef2Dirty.keySet().contains(libRef)) {
			return libRef2Dirty.get(libRef);
		}
		libRef2Dirty.put(libRef, true);
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public void setDirty(ILibraryElementReference libRef, boolean flag) {
		libRef2Dirty.put(libRef, flag);
	}

	/** Registers a changes observer on the library element. */
	protected void registerChangeObserverOnLibraryElement(final ILibraryElementReference ref,
			ILibraryElement libraryElement) {
		EContentAdapter adapter = new EContentAdapter() {
			@Override
			public void notifyChanged(Notification notification) {
				super.notifyChanged(notification);
				setDirty(ref, true);
			}
		};
		libraryElement.eAdapters().add(adapter);
	}

	/** {@inheritDoc} */
	@Override
	public ILibraryElement findLibraryElementByURI(String uri) {
		for(ILibraryElement libElem : getLibraryElementsFromAllWorkspaceLibraries()) {
			if(libElem.getURI().equals(uri)) {
				return libElem;
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public List<ILibraryElement> getLibraryElementsFromAllWorkspaceLibraries() {
		EList<ILibraryElement> libraryElements = new BasicEList<ILibraryElement>();
		for(ILibrary lib : getLibrariesFromWorkspace()) {
			for(ILibraryPackage lcr : getAllLibraryPackages(lib)) {
				libraryElements.addAll(lcr.getLibraryElements());
			}
		}
		return libraryElements;
	}

	/** {@inheritDoc} */
	@Override
	public List<ILibrary> getLibrariesFromWorkspace() {
		List<ILibrary> libs = new LinkedList<ILibrary>();

		List<ITopLevelElement> topLevelElements =
				IPersistencyService.INSTANCE.getTopLevelElements();
		for(ITopLevelElement current : topLevelElements) {
			EObject root = current.getRootModelElement();
			if(root instanceof ILibrary) {
				libs.add((ILibrary)root);
			}
		}

		return libs;
	}

	/** {@inheritDoc} */
	@Override
	public List<ILibraryPackage> getAllLibraryPackages(ILibrary lib) {
		List<ILibraryPackage> result = new ArrayList<ILibraryPackage>();

		for(ILibraryPackage topLevelPack : pickInstanceOf(ILibraryPackage.class,
				lib.getRootElements())) {
			result.add(topLevelPack);
			result.addAll(getAllSubPackages(topLevelPack));
		}

		return result;
	}

	/** All sub-packages. */
	public List<ILibraryPackage> getAllSubPackages(ILibraryPackage pack) {
		List<ILibraryPackage> result = new ArrayList<ILibraryPackage>();

		for(ILibraryPackage subPack : pack.getSubPackages()) {
			result.add(subPack);
			result.addAll(getAllSubPackages(subPack));
		}

		return result;
	}

	/**
	 * The set of component references for which change adapters on the corresponding library
	 * objects were set.
	 */
	private HashSet<ILibraryElementReference> libraryElementsReferencesWithRegisteredAdapters =
			new HashSet<ILibraryElementReference>();

	/** {@inheritDoc} */
	@Override
	public void ensureWrappedElementIsUpToDate(final ILibraryElementReference ref) {
		if(!isDirty(ref)) {
			return;
		}

		setDirty(ref, false);

		final ILibraryElement lc = findLibraryElementByURI(ref.getURI());
		if(lc == null) {
			return;
		}

		Runnable cmd = new Runnable() {
			@Override
			public void run() {
				ILibraryService.INSTANCE.setDirty(ref, false);
				EObject copy = copy(lc.getWrappedElement());
				ref.setLibraryElementShadow(copy);

				// create new IDs
				UniqueIDUtils.prepareUniqueID(copy, ref);
			}
		};

		ITopLevelElement topLevel = IPersistencyService.INSTANCE.getTopLevelElementFor(ref);
		if(topLevel != null) {
			topLevel.runAsNonDirtyingCommand(cmd);
		} else {
			cmd.run();
		}

		if(!libraryElementsReferencesWithRegisteredAdapters.contains(ref)) {
			registerChangeObserverOnLibraryElement(ref, lc);
			libraryElementsReferencesWithRegisteredAdapters.add(ref);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void ensureChildrenLibraryReferenceAreUpToDate(EObject context) {
		EList<ILibraryElementReference> libReferences;
		libReferences = getChildrenWithType(context, ILibraryElementReference.class);
		for(ILibraryElementReference ref : libReferences) {
			ensureWrappedElementIsUpToDate(ref);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void addElementToLibrary(final ILibraryPackage pack, final EObject namedElement) {
		ITopLevelElement topLevel = IPersistencyService.INSTANCE.getTopLevelElementFor(pack);
		topLevel.runAsCommand(new Runnable() {
			@Override
			public void run() {
				EObject copy = copy(namedElement);

				Class<? extends EObject> clazz = namedElement.getClass();
				List<ILibraryElementHandler<EObject>> handlers = getRegisteredHandlers(clazz);
				assert handlers.size() == 1;

				ILibraryElementHandler<EObject> handler = handlers.get(0);
				ILibraryElement libElem = handler.createLibraryElement(copy);
				pack.getLibraryElements().add(libElem);

				for(IPrototypeProvider pp : IPrototypeService.INSTANCE.getPrototypeProviders()) {
					if(pp instanceof LibraryPrototypeProvider) {
						((LibraryPrototypeProvider)pp).refreshPrototypes();
					}
				}
			}
		});
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}
}
