/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.swt;

import java.awt.image.BufferedImage;
import java.awt.image.DirectColorModel;
import java.awt.image.WritableRaster;

import org.conqat.lib.commons.assertion.CCSMAssert;
import org.conqat.lib.commons.string.StringUtils;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Layout;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

/**
 * Provides a bunch of commonly used SWT constructs.
 * 
 * @author $Author: hummelb $
 * @version $Rev: 36363 $
 * @ConQAT.Rating YELLOW Hash: A352E3F352E399BC63F62B7BF569A85A
 */
public class SWTUtils {

	/**
	 * Creates and returns a {@link Label} in the given parent with the given
	 * text.
	 */
	public static Label createLabel(Composite parent, String text) {
		Label label = new Label(parent, SWT.NONE);
		label.setText(text);
		layout(label);
		return label;
	}

	/**
	 * Creates and returns a single line {@link Text} editor in the given parent
	 * initially with the given default text and with the given {@link Listener}
	 * set as {@link SWT#Modify} listener.
	 * <p>
	 * If the modify listener is <code>null</code> it is not added.
	 */
	public static Text createText(Composite parent, String defaultText,
			Listener modifyListener) {
		Text text = new Text(parent, SWT.BORDER);
		text.setText(defaultText);
		if (modifyListener != null) {
			text.addListener(SWT.Modify, modifyListener);
		}
		layout(text);
		return text;
	}

	/**
	 * Creates and returns a {@link Button} in the given parent with the given
	 * text, the given style and with the given selection listener. A default
	 * (grid) layout is applied to the button.
	 * <p>
	 * If the selection listener is <code>null</code> it is not added.
	 */
	public static Button createButton(Composite parent, String text, int style,
			SelectionListener selectionListener) {
		Button button = new Button(parent, style);
		button.setText(text);
		if (selectionListener != null) {
			button.addSelectionListener(selectionListener);
		}
		layout(button);
		return button;
	}

	/** Creates a {@link Group}. */
	public static Group createGroup(Composite parent, String text, Layout layout) {
		Group group = new Group(parent, SWT.NONE);
		group.setText(text);
		group.setLayout(layout);
		SWTUtils.layout(group);
		return group;
	}

	/** Layouts a control using default settings of grid layout. */
	public static void layout(Control control) {
		GridDataFactory.defaultsFor(control).applyTo(control);
	}

	/** Convert BufferedImage to SWT ImageData */
	public static ImageData convertToSWT(BufferedImage bufferedImage) {
		boolean typeCheck = bufferedImage.getColorModel() instanceof DirectColorModel;
		CCSMAssert.isTrue(typeCheck, "The buffered image must use "
				+ "DirectColorModel");

		DirectColorModel colorModel = (DirectColorModel) bufferedImage
				.getColorModel();
		PaletteData palette = new PaletteData(colorModel.getRedMask(),
				colorModel.getGreenMask(), colorModel.getBlueMask());
		ImageData data = new ImageData(bufferedImage.getWidth(),
				bufferedImage.getHeight(), colorModel.getPixelSize(), palette);
		WritableRaster raster = bufferedImage.getRaster();

		int[] pixelArray = new int[4];
		for (int y = 0; y < data.height; y++) {
			for (int x = 0; x < data.width; x++) {
				raster.getPixel(x, y, pixelArray);
				int pixel = palette.getPixel(new RGB(pixelArray[0],
						pixelArray[1], pixelArray[2]));
				data.setPixel(x, y, pixel);
			}
		}

		return data;
	}

	/** Shows the busy cursor on the active shell of the default display. */
	public static void showBusyCursor() {
		Shell shell = Display.getDefault().getActiveShell();
		if (shell == null) {
			return;
		}
		shell.setCursor(Display.getDefault().getSystemCursor(SWT.CURSOR_WAIT));
	}

	/** Shows the default cursor on the active shell of the default display. */
	public static void showDefaultCursor() {
		Shell shell = Display.getDefault().getActiveShell();
		if (shell == null) {
			return;
		}
		shell.setCursor(null);
	}

	/** Creates an empty label as separator. */
	public static void createSeparator(Composite parent) {
		createLabel(parent, StringUtils.EMPTY_STRING);
	}
}