/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.lib.commons.string;

import java.util.Arrays;

import org.conqat.lib.commons.assertion.CCSMPre;
import org.conqat.lib.commons.assertion.PreconditionException;
import org.conqat.lib.commons.collections.ManagedIntArray;

/**
 * A class that helps to convert between line numbers and character offsets. The
 * character offset is zero based while the line number is one based.
 * <p>
 * Note that this class only works for strings where the line terminator is
 * '\n'. Other strings have to be converted first (e.g. using
 * {@link StringUtils#replaceLineBreaks(String, String)}).
 * <p>
 * The implementation works by storing all character offsets of the newline
 * characters. For space/performance reasons, these offsets are stored in a bare
 * array, which is managed by the base class.
 * <p>
 * <i>Note:</i> This class extends {@link ManagedIntArray} although delegation
 * would be more elegant. The problem, however, is, that {@link ManagedIntArray}
 * does not support delegation. The idea of this class was to provide protected
 * access to its internals, which only works via inheritance. Actually, this is
 * not interface inheritance ({@link ManagedIntArray} has no public method,
 * btw.) but implementation inheritance.
 * 
 * @author $Author: deissenb $
 * @version $Rev: 35183 $
 * @ConQAT.Rating GREEN Hash: CD6A4DF7B098312F97E95F9E0F08F859
 */
public class LineOffsetConverter extends ManagedIntArray {

	/**
	 * Creates a converter for a specific string.
	 * 
	 * @param s
	 *            the string for which conversion should be provided. This only
	 *            works, if the string contains only '\n' as line terminator. If
	 *            this is not the case, the
	 *            {@link StringUtils#replaceLineBreaks(String, String)} method
	 *            can be useful.
	 * 
	 * @throws PreconditionException
	 *             if the string contains a '\r'.
	 */
	public LineOffsetConverter(String s) {
		char[] chars = s.toCharArray();
		for (int i = 0; i < chars.length; ++i) {
			if (chars[i] == '\n') {
				addArrayElement();
				array[size - 1] = i;
			} else if (chars[i] == '\r') {
				throw new PreconditionException(
						"The string may not contain '\\r', but does so at position "
								+ i);
			}
		}
	}

	/**
	 * Returns the (zero based) offset of the first character of the given line
	 * (starting at 1).
	 * 
	 * @throws PreconditionException
	 *             if the line is not valid for the string.
	 */
	public int getOffset(int line) {
		CCSMPre.isTrue(1 <= line && line <= size + 1,
				"This is not a valid line: " + line);

		if (line == 1) {
			return 0;
		}

		// first character of line is directly behind newline character
		return array[line - 2] + 1;
	}

	/**
	 * Returns the (one based) line for the character at the given (zero based)
	 * offset. The newline at the end of a line is counted as a part of the line
	 * (i.e. for the very first newline, we would return 1). If the offset is
	 * larger than the length of the string, the index of the last line is
	 * returned.
	 */
	public int getLine(int offset) {
		CCSMPre.isTrue(offset >= 0, "Negative offsets not supported!");

		int index = Arrays.binarySearch(array, 0, size, offset);
		if (index >= 0) {
			return index + 1;
		}

		return -index;
	}
}